@@ -9,12 +9,13 @@ from __future__ import absolute_import
 
 import os
 import sys
+import math
 import random
 
 from bitarray import bitarray, bits2bytes
 
 from bitarray._util import (
-    zeros, ones, count_n, parity, xor_indices,
+    zeros, ones, count_n, parity, sum_indices, xor_indices,
     count_and, count_or, count_xor, any_and, subset,
     correspond_all, byteswap,
     serialize, deserialize,
@@ -26,9 +27,9 @@ from bitarray._util import (
 )
 
 __all__ = [
-    'zeros', 'ones', 'urandom', 'random_p',
+    'zeros', 'ones', 'urandom', 'random_k', 'random_p',
     'pprint', 'strip', 'count_n',
-    'parity', 'xor_indices',
+    'parity', 'sum_indices', 'xor_indices',
     'count_and', 'count_or', 'count_xor', 'any_and', 'subset',
     'correspond_all', 'byteswap', 'intervals',
     'ba2hex', 'hex2ba',
@@ -51,13 +52,34 @@ Return random bitarray of length `n` (uses `os.urandom()`).
     return a
 
 
+def random_k(__n, k, endian=None):
+    """random_k(n, /, k, endian=None) -> bitarray
+
+Return (pseudo-) random bitarray of length `n` with `k` elements
+set to one.  Mathematically equivalent to setting (in a bitarray of
+length `n`) all bits at indices `random.sample(range(n), k)` to one.
+The random bitarrays are reproducible when giving Python's `random.seed()`
+with a specific seed value.
+
+This function requires Python 3.9 or higher, as it depends on the standard
+library function `random.randbytes()`.  Raises `NotImplementedError`
+when Python version is too low.
+"""
+    if sys.version_info[:2] < (3, 9):
+        raise NotImplementedError("bitarray.util.random_k() requires "
+                                  "Python 3.9 or higher")
+    r = _Random(__n, endian)
+    return r.random_k(k)
+
+
 def random_p(__n, p=0.5, endian=None):
     """random_p(n, /, p=0.5, endian=None) -> bitarray
 
-Return pseudo-random bitarray of length `n`.  Each bit has probability `p` of
-being 1.  Equivalent to `bitarray((random() < p for _ in range(n)), endian)`.
-The bitarrays are reproducible when calling Python's `random.seed()` with a
-specific seed value.
+Return (pseudo-) random bitarray of length `n`, where each bit has
+probability `p` of being one (independent of any other bits).  Mathematically
+equivalent to `bitarray((random() < p for _ in range(n)), endian)`, but much
+faster for large `n`.  The random bitarrays are reproducible when giving
+Python's `random.seed()` with a specific seed value.
 
 This function requires Python 3.12 or higher, as it depends on the standard
 library function `random.binomialvariate()`.  Raises `NotImplementedError`
@@ -66,10 +88,16 @@ when Python version is too low.
     if sys.version_info[:2] < (3, 12):
         raise NotImplementedError("bitarray.util.random_p() requires "
                                   "Python 3.12 or higher")
-    r = _RandomP(__n, endian)
+    r = _Random(__n, endian)
     return r.random_p(p)
 
-class _RandomP:
+class _Random:
+
+    # The main reason for this class it to enable testing functionality
+    # individually in the test class Random_P_Tests in 'test_util.py'.
+    # The test class also contains many comments and explanations.
+    # To better understand how the algorithm works, see ./doc/random_p.rst
+    # See also, VerificationTests in ./devel/test_random.py
 
     # maximal number of calls to .random_half() in .combine()
     M = 8
@@ -94,47 +122,79 @@ class _RandomP:
         del a[self.n:]
         return a
 
-    def get_op_seq(self, i):
+    def op_seq(self, i):
         """
         Return bitarray containing operator sequence.
-        Each item represents a bitwise operation:   0: and   1: or
-        After applying the sequence (see .random_combine()), we
+        Each item represents a bitwise operation:   0: AND   1: OR
+        After applying the sequence (see .combine_half()), we
         obtain a bitarray with probability  q = i / K
         """
-        assert 0 < i < self.K
-        # sequence of &, | operations - least significant operations first
-        seq = int2ba(i, length=self.M, endian="little")
-        seq = strip(seq, mode="left")
-        del seq[0]
-        return seq
+        if not 0 < i < self.K:
+            raise ValueError("0 < i < %d, got i = %d" % (self.K, i))
 
-    def random_combine(self, p):
+        # sequence of &, | operations - least significant operations first
+        a = bitarray(i.to_bytes(2, byteorder="little"), "little")
+        return a[a.index(1) + 1 : self.M]
+
+    def combine_half(self, seq):
         """
-        Given a desired probability p, return a random bitarray with
-        actual probability  q = int(p * K) / K
+        Combine random bitarrays with probability 1/2
+        according to given operator sequence.
         """
         a = self.random_half()
-        for k in self.get_op_seq(int(p * self.K)):
-            b = self.random_half()
+        for k in seq:
             if k:
-                a |= b
+                a |= self.random_half()
             else:
-                a &= b
+                a &= self.random_half()
         return a
 
-    def random_m(self, m):
-        """
-        Return a bitarray (of length n) with m bits randomly set to 1.
-        Designed for small m (compared to n).
-        """
-        assert 0 <= m <= self.n
-        a = zeros(self.n, self.endian)
-        for _ in range(m):
-            while 1:
-                i = random.randrange(self.n)
-                if not a[i]:
-                    a[i] = 1
-                    break
+    def random_k(self, k):
+        n = self.n
+        # error check inputs and handle edge cases
+        if k <= 0 or k >= n:
+            if k == 0:
+                return zeros(n, self.endian)
+            if k == n:
+                return ones(n, self.endian)
+            raise ValueError("k must be in range 0 <= k <= n, got %s" % k)
+
+        # exploit symmetry to establish: k <= n // 2
+        if k > n // 2:
+            a = self.random_k(n - k)
+            a.invert()  # use in-place to avoid copying
+            return a
+
+        # decide on sequence, see VerificationTests ./devel/test_random.py
+        if k < 16 or k * self.K < 3 * n:
+            i = 0
+        else:
+            p = k / n  # p <= 0.5
+            p -= (0.2 - 0.4 * p) / math.sqrt(n)
+            i = int(p * (self.K + 1))
+
+        # combine random bitarrays using bitwise AND and OR operations
+        if i < 3:
+            a = zeros(n, self.endian)
+            diff = -k
+        else:
+            a = self.combine_half(self.op_seq(i))
+            diff = a.count() - k
+
+        randrange = random.randrange
+        if diff < 0:  # not enough bits 1 - increase count
+            for _ in range(-diff):
+                i = randrange(n)
+                while a[i]:
+                    i = randrange(n)
+                a[i] = 1
+        elif diff > 0:  # too many bits 1 - decrease count
+            for _ in range(diff):
+                i = randrange(n)
+                while not a[i]:
+                    i = randrange(n)
+                a[i] = 0
+
         return a
 
     def random_p(self, p):
@@ -149,27 +209,40 @@ class _RandomP:
             raise ValueError("p must be in range 0.0 <= p <= 1.0, got %s" % p)
 
         # for small n, use literal definition
-        if self.n < 10:
+        if self.n < 16:
             return bitarray((random.random() < p for _ in range(self.n)),
                             self.endian)
 
-        # exploit symmetry to establish: p <= 0.5
+        # exploit symmetry to establish: p < 0.5
         if p > 0.5:
             a = self.random_p(1.0 - p)
-            a.invert()
+            a.invert()  # use in-place to avoid copying
             return a
 
         # for small p, set randomly individual bits
         if p < self.SMALL_P:
-            return self.random_m(random.binomialvariate(self.n, p))
+            return self.random_k(random.binomialvariate(self.n, p))
 
-        # combine random bitarrays using bitwise & and | operations
-        a = self.random_combine(p)
-        q = int(p * self.K) / self.K
+        # calculate operator sequence
+        i = int(p * self.K)
+        if p * (self.K + 1) > i + 1: # see ./devel/test_random.py
+            i += 1
+        seq = self.op_seq(i)
+        q = i / self.K
+
+        # when n is small compared to number of operations, also use literal
+        if self.n < 100 and self.nbytes <= len(seq) + 3 * bool(q != p):
+            return bitarray((random.random() < p for _ in range(self.n)),
+                            self.endian)
+
+        # combine random bitarrays using bitwise AND and OR operations
+        a = self.combine_half(seq)
         if q < p:
-            # increase probability q by "oring" with probability x
             x = (p - q) / (1.0 - q)
             a |= self.random_p(x)
+        elif q > p:
+            x = p / q
+            a &= self.random_p(x)
 
         return a
 
